// **********************************************************************************
// Internet Of Things Client Framework
//
// This has all the code for the system
// Remember, you need to code the following functions if you are using the generic
// version of this software:
//
// uartSend()  <- This is how your system sends stuff to uart
// uartInit()  <- This is how your system initialises the uart module
// uartGet()   <- Just remember to call this upon character receive!
// **********************************************************************************
	
	#include <stdlib.h>
	#include "IoT_ClientFramework.h"

	
	// ******************************************************************************
	// Enter your device header here
	// ******************************************************************************
		#include <p18f25k20.h>
	// ******************************************************************************




	// ******************************************************************************
	// Please define these
	// ******************************************************************************
	

		// Function		: uartSend(byteToSend)
		// Description	: Sends a byte over UART
		// Returns		: None	
		void uartSend(unsigned char byteToSend)
		{
			TXREG = byteToSend;				// Transfer byte to UART send register
			while(!TXSTAbits.TRMT);			// Wait until byte sent
		}
		

		
		// Function		: uartInit()
		// Description	: Configures UART to be 9600 baud
		// Returns		: None
		void uartInit(void)
		{
			TXSTAbits.BRGH = 1;				// High Baud Rate
			BAUDCONbits.BRG16 = 1;			// 16 Bit SPBRG
			BAUDCONbits.CKTXP = 0;			// Idle on high
			SPBRGH = 0x01;
			SPBRG = 0xA0;	
			TXSTAbits.SYNC = 0;				// Async comms
			RCSTAbits.SPEN = 1;				// Enable serial port
			TXSTAbits.TXEN = 1;				// Enable transmission
			RCSTAbits.CREN = 1;				// Enable reception
			PIE1bits.RCIE = 1;				// Enable reception interrupt
		}
		
		
		
		// Function		: delay20ms()
		// Description	: Make this function take 20ms to complete
		// Returns		: None
		void delay20ms(void)
		{
			unsigned long double x;
			for(x = 0; x < 500; x ++)
			{
			}
		}
	
	
	
	// ******************************************************************************
	// Constants
	// ******************************************************************************
		rom const char instructionArray[][18] = 		
		{
			{"OK"},
			{"SEND OK"},
			{"SEND FAIL"},
			{"SEND ERROR"},
			{"ERROR"},
			{"FAIL"},
			{"busy p..."},
			{"busy s..."},
			{"+IPD"},
			{"> "},
			{"CONNECT"},
			{"CLOSED"},
			{"WIFI CONNECTED"},
			{"WIFI DISCONNECTED"}
		};
		
		rom const char at[]		 	= "AT";
		rom const char echoOff[] 	= "ATE0";
		rom const char wifiCon[] 	= "AT+CWJAP=";
		rom const char cwmode[]		= "AT+CWMODE=1";
		rom const char cipmux[]	 	= "AT+CIPMUX=0";
		rom const char cipstart[] 	= "AT+CIPSTART=";
		rom const char baudChange[]	= "AT+IPR=9600";
		rom const char cipsend[]	= "AT+CIPSEND=";
		rom const char tcp[]		= "TCP";
	// ******************************************************************************

	
	
	// ******************************************************************************
	// RAM Variables
	// ******************************************************************************
		
		unsigned char instructionStates[NO_INSTRUCTIONS];
		unsigned char uartData;
		unsigned char uartNewData;
			
		unsigned char dataLength;
		unsigned char dataSendLength;
		char  *dataBuffPtr;
		unsigned char connectionStatus;
		unsigned char wifiStatus;
		unsigned char dataAvailable;
		unsigned char dataPtr;
		char dataBuffer[32];
		char dataSizeText[3];
		char dataLengthText[3];
		char deviceID[1] = {0};
	
		unsigned char commandInProgress;
		unsigned char commandSuccess;
		unsigned char strPointer;
		unsigned char responseNumber;
		unsigned char commandMode;
		unsigned char dataMode;	
		unsigned char ipdMode;
		unsigned char validateMode;
		unsigned char sendDataNow;
		
		unsigned char riotServerMode;
		unsigned char riotDataType;
		unsigned char riotDataHandle;
		
		unsigned char connectReceived;
		unsigned char closedReceived;
		
		unsigned char handleComplete;
		
		unsigned int sendTimer;
	// ******************************************************************************
	



// ==================================================================================
// User Commands
// ==================================================================================

	// ******************************************************************************
	// Function		: internet_init()
	// Description	: Sets up the ESP8266 and the framework core
	// Returns		: TRUE if success, FALSE if something fatal happened
	// ******************************************************************************	
	unsigned char internet_init(void)
	{
		unsigned char cnt;
		unsigned char tempPtr = 0;
		unsigned long int x;
						
		uartInit();
		connectionStatus 	= FALSE;
		wifiStatus 			= FALSE;
		validateMode		= TRUE;
		commandMode 		= FALSE;
		commandInProgress 	= FALSE;
		uartData 			= FALSE;
		uartNewData	 		= FALSE;
		handleComplete 		= FALSE;
		sendDataNow			= FALSE;
		ipdMode				= FALSE;
		riotServerMode		= FALSE;
		riotDataHandle		= FALSE;
		dataLength 			= 0;
		dataPtr				= 0;
		strPointer			= 0;


		
		for(cnt = 0; cnt < NO_INSTRUCTIONS; cnt ++)
		{
			instructionStates[cnt] = TRUE;
		}
		
		for(cnt = 0; cnt < 100; cnt ++)
		{
			delay20ms();
		}

		commandSend(at);
		internet_waitUntilDone();
		
		commandSend(echoOff);
		internet_waitUntilDone();
		
		commandSend(cwmode);
		internet_waitUntilDone();
		
		commandSend(cipmux);	
		internet_waitUntilDone();
	}
	// ******************************************************************************



	// ******************************************************************************
	// Function		: internet_connectToWifi(SSID, PASSWORD)
	// Description	: Connect to a WiFi network
	// Returns		: TRUE if command accepted, FALSE if command engine busy
	// ******************************************************************************	
	unsigned char internet_connectToWifi(rom const char *wifiName,rom const char *wifiPassword)
	{
		unsigned char cnt; 
		delay20ms();
				
		if(commandInProgress == FALSE)
		{		
			commandInProgress = TRUE;
			
			cnt = 0;
			while(wifiCon[cnt] != 0x00)
			{
				uartSend(wifiCon[cnt]);
				cnt ++;
			}
			uartSend(SPEECH);
			cnt = 0;
			
			while(wifiName[cnt] != 0x00)
			{
				uartSend(wifiName[cnt]);
				cnt ++;
			}						
			uartSend(SPEECH);
			uartSend(COMMA);
			uartSend(SPEECH);
			cnt = 0;
			
			while(wifiPassword[cnt] != 0x00)
			{
				uartSend(wifiPassword[cnt]);
				cnt ++;
			}						
			uartSend(SPEECH);
			uartSend(CR);
			uartSend(LF);
			
			internet_waitUntilDone();
			
			// If the connection fails, respond with fail!
			if(commandSuccess == FALSE)
			{
				return FALSE;
			}
	
			return TRUE;	
		}
		else
		{
			return FALSE;
		}
	}
	// ******************************************************************************
	
	
		
	// ******************************************************************************
	// Function		: internet_waitUntilDone()
	// Description	: When called will loop until a command ack is received
	// Returns		: None
	// ******************************************************************************
	unsigned char internet_waitUntilDone(void)
	{
		while(commandInProgress == TRUE)
		{
			internet_update();
		}	
	}
	// ******************************************************************************
	
	
	
	// ******************************************************************************
	// Function		: internet_connectToServer(IP, PORT)
	// Description	: Connects to a server defined by "ip" on port "port"
	// Returns		: FALSE if command engine is busy or failed
	// ******************************************************************************	
	unsigned char internet_connectToRIOTServer(rom const char *ip, rom const char *port)
	{
		unsigned char cnt; 
		delay20ms();
			
		if(commandInProgress == FALSE)
		{		
			commandInProgress = TRUE;
			
			cnt = 0;
			while(cipstart[cnt] != 0x00)
			{
				uartSend(cipstart[cnt]);
				cnt ++;
			}
			uartSend(SPEECH);
			cnt = 0;
			
			while(tcp[cnt] != 0x00)
			{
				uartSend(tcp[cnt]);
				cnt ++;
			}						
			uartSend(SPEECH);
			uartSend(COMMA);
			uartSend(SPEECH);
			cnt = 0;
			
			while(ip[cnt] != 0x00)
			{
				uartSend(ip[cnt]);
				cnt ++;
			}						
			uartSend(SPEECH);
			uartSend(COMMA);
			
			cnt = 0;
			
			while(port[cnt] != 0x00)
			{
				uartSend(port[cnt]);
				cnt ++;
			}						
			
			
			uartSend(CR);
			uartSend(LF);
			
			internet_waitUntilDone();

			if(commandSuccess == FALSE)
			{
				return FALSE;
			}
			
			// Wait until the dynamically allocated port is sent
			while(dataAvailable == FALSE)
			{
			}
			
			// Add a 0x00 terminator to the end of the data
			dataBuffer[dataLength] = 0x00;
			
			cnt = 0;
			while(cipstart[cnt] != 0x00)
			{
				uartSend(cipstart[cnt]);
				cnt ++;
			}
			uartSend(SPEECH);
			cnt = 0;
			
			while(tcp[cnt] != 0x00)
			{
				uartSend(tcp[cnt]);
				cnt ++;
			}						
			uartSend(SPEECH);
			uartSend(COMMA);
			uartSend(SPEECH);
			cnt = 0;
			
			while(ip[cnt] != 0x00)
			{
				uartSend(ip[cnt]);
				cnt ++;
			}						
			uartSend(SPEECH);
			uartSend(COMMA);
			
			cnt = 0;
			
			while(dataBuffer[cnt] != 0x00)
			{
				uartSend(dataBuffer[cnt]);
				cnt ++;
			}						
			
			
			uartSend(CR);
			uartSend(LF);
			
			internet_waitUntilDone();			
			
			// If the connection fails, respond with fail!
			if(commandSuccess == FALSE)
			{
				return FALSE;
			}
						
			riotServerMode = TRUE;
			
			for(cnt = 0; cnt < 20; cnt ++)
			{
				delay20ms();
			}
						
			return TRUE;	
		}
		else
		{
			return FALSE;
		}		
	}
	// ******************************************************************************	
	


	// ******************************************************************************
	// Function		: internet_sendData(char *data, char length)
	// Description	: Sends data pointed by data to connected server
	// Returns		: FALSE if command engine is busy
	// ******************************************************************************	
	unsigned char internet_sendData(char *data, char length)
	{
		// Step 1 - Send AT+CIPSEND=length
		// Step 2 - Wait for >
		// Step 3 - Send the data
		unsigned char cnt;
		
		if(commandInProgress == FALSE)
		{		
			commandInProgress = TRUE;

			if(riotServerMode == TRUE)
			{
				length ++;	
			}

			itoa(length, dataLengthText);
			dataSendLength = length;
			dataBuffPtr = data;
			
			
			cnt = 0;
			
		
			while(cipsend[cnt] != 0x00)
			{
				uartSend(cipsend[cnt]);
				cnt ++;
			}			
			
			cnt = 0;
			
			while(dataLengthText[cnt] != 0x00)
			{
				uartSend(dataLengthText[cnt]);
				cnt ++;
			}	
			uartSend(CR);
			uartSend(LF);			
									
			return TRUE;	
		}
		else
		{
			return FALSE;
		}			
	}
	// ******************************************************************************
	


	// ******************************************************************************
	// Function		: internet_newDataAvailable(void)
	// Description	: Informs user if new data available
	// Returns		: FALSE if no new data
	// ******************************************************************************
	unsigned char internet_newDataAvailable(void)
	{
		if(dataAvailable == TRUE)
		{
			dataAvailable = FALSE;
			return TRUE;
		}
		else
		{
			return FALSE;
		}
	}
	// ******************************************************************************	
	
	

	// ******************************************************************************
	// Function		: internet_getDataLength(void)
	// Description	: Informs user if new data available
	// Returns		: FALSE if no new data
	// ******************************************************************************	
	unsigned char internet_getDataLength(void)
	{
		return dataLength;	
	}	
	// ******************************************************************************		



	// ******************************************************************************
	// Function		: internet_getDataLength(void)
	// Description	: Informs user if new data available
	// Returns		: FALSE if no new data
	// ******************************************************************************	
	char *internet_getNewData(void)
	{
		return dataBuffer;	
	}	
	// ******************************************************************************


	
	// ******************************************************************************
	// Function		: internet_connectionStatus(void)
	// Description	: Returns the status of the connection in proress
	// Returns		: FALSE if not connected
	// ******************************************************************************	
	unsigned char internet_connectionStatus(void)
	{
		return connectionStatus;
	}	
	// ******************************************************************************
		


	// ******************************************************************************
	// Function		: internet_wifiStatus(void)
	// Description	: Returns the status of the wifi connection
	// Returns		: FALSE if not connected
	// ******************************************************************************	
	unsigned char internet_wifiStatus(void)
	{
		return wifiStatus;
	}	
	// ******************************************************************************

	
		
	// ******************************************************************************
	// Function		: internet_setdeviceID(unsigned char id)
	// Description	: Sets the unique ID of the device (256 possibilities)
	// Returns		: None
	// ******************************************************************************
	unsigned char internet_setdeviceID(unsigned char id)
	{
		deviceID[0] = id;	
		riotDataType = RIOTDEVID;	
		
		if(internet_sendData(deviceID,1))
		{
			internet_waitUntilDone();
			return TRUE;
		}
		else
		{
			return FALSE;
		}
	}
	// ******************************************************************************
	
	

	// ******************************************************************************
	// Function		: internet_setdeviceID(unsigned char id)
	// Description	: Sets the unique ID of the device (256 possibilities)
	// Returns		: None
	// ******************************************************************************
	unsigned char riot_sendData(char *data, char length)
	{
		riotDataType = RIOTDATA;	
		
		if(internet_sendData(data, length))
		{
			return TRUE;
		}
		else
		{
			return FALSE;
		}
	}
	// ******************************************************************************

	
	
		
// ==================================================================================




// **********************************************************************************
// FRAMEWORK CODE - HERE BE DRAGONS
// **********************************************************************************
	
	
	// ******************************************************************************
	// Function		: internet_update()
	// Description	: Framework Core - Updates as fast as possible
	// Returns		: None
	// ******************************************************************************	
	void internet_update(void)
	{			
		if(validateMode == TRUE)
		{	
			if(uartNewData == TRUE)
			{
				uartNewData = FALSE;
				if(commandValidate() == TRUE)
				{
					// Decide which mode to go into
					if(sendDataNow == TRUE)
					{
						delay20ms();
						
						if(riotServerMode == TRUE)
						{
							dataSendLength --;
							
							switch(riotDataType)
							{
								case RIOTDEVID:
									uartSend('I');
									break;
									
								case 1:
									uartSend('D');
									break;
									
								case 2:
									uartSend('R');
									break;
									
								case RIOTTIMEOUT:
									uartSend('A');
									break;
							}
						}
						
						for(dataPtr = 0; dataPtr < dataSendLength; dataPtr++)
						{
							uartSend(dataBuffPtr[dataPtr]);
						}
						
						sendTimer = 0;
						handleComplete = TRUE;						
					}
					else
					{
						commandMode = TRUE;
					}	
					
					validateMode = FALSE;	
				}	
			}					
		}
		else if(commandMode == TRUE)
		{
			// Command mode processes the accepted command
			commandHandle();
			tcpHandle();
			handleComplete = TRUE;
		}	
		else if(ipdMode == TRUE)
		{			
			if(uartNewData == TRUE)
			{
				// Data mode takes the raw byte in and adds to the data buffer

				uartNewData = FALSE;
				dataHandle();
			}
		}
		else
		{
			// Command is nothing and probably invalid so lets
			// clean everything up and start from scratch
			handleComplete = TRUE;
		}
		
		if(handleComplete == TRUE)
		{
			timeoutChecker();
			cleanup();
		}			
	}
	// ******************************************************************************
	
	
	
	
	// ******************************************************************************
	// Function		: commandSend(string)
	// Description	: Sends a command to the ESP8266
	// Returns		: TRUE if the command was accepted. FALSE if system busy
	// ******************************************************************************
	unsigned char commandSend(rom const char *string)
	{
		if(commandInProgress == FALSE)
		{
			unsigned char cnt = 0x00; 
						
			commandInProgress = TRUE;
			
			while(string[cnt] != 0x00)
			{
				uartSend(string[cnt]);
				cnt++;
			}
			uartSend(CR);
			uartSend(LF);	
			
			return TRUE;	
		}
		else
		{
			return FALSE;
		}
	}
	
	// ******************************************************************************




	// ******************************************************************************
	// Function		: commandValidate
	// Description	: Validates the command that has been sent from ESP8266
	// Returns		: TRUE if a command has been sent. FALSE if waiting for more data
	// ******************************************************************************
	unsigned char commandValidate(void)
	{
		unsigned char cnt;
		
		// Is this a carriage return?
		if(uartData == CR)
		{
			responseNumber = 0xFF;
			
			for(cnt = 0; cnt < NO_INSTRUCTIONS; cnt ++)
			{
				// Skip all not equal instructions
				if(instructionStates[cnt] == TRUE)
				{
					// If a string has 0x00 the thats the instruction
					if(instructionArray[cnt][strPointer] != 0x00)
					{
						instructionStates[cnt] = FALSE;
					}
					else
					{
						responseNumber = cnt;
					}					
				}
			}
			
			return TRUE;
		}
		
		// In command mode, Line Feeds are skipped
		if(uartData == LF)
		{
			return FALSE;
		}
		
		// Ignore white space in the start
		// Also, > is a high priority
		
		if(strPointer == 0)
		{
			if(uartData == SPACE)
			{
				return FALSE;
			}
			else if(uartData == PROMPTSYM)
			{
				sendDataNow = TRUE;
				return TRUE;
			}	
		}
		
		
		// Is the string +IPD?
		if(strPointer == 3)
		{
			if(instructionStates[IPD] == TRUE)
			{
				validateMode = FALSE;
				ipdMode = TRUE;
				dataMode = FALSE;
				return FALSE;
			}
		}
		

		// Firstly, go through all command chars and avoid mismatches
		for(cnt = 0; cnt < NO_INSTRUCTIONS; cnt ++)
		{
			// Skip all arrays that have been confirmed not related
			if(instructionStates[cnt] == TRUE)
			{	
				if(instructionArray[cnt][strPointer] != uartData)
				{
					instructionStates[cnt] = FALSE;
				}
			}
		}
		
		strPointer ++;
		return FALSE;
	}
	// ******************************************************************************




	// ******************************************************************************
	// Function		: dataHandle
	// Description	: Receives incoming data and adds to buffer
	// Returns		: None
	// ******************************************************************************
	void dataHandle(void)
	{
		// Firstly, we have to get the data size!
		if(dataMode == FALSE)
		{
			if(uartData == SEMICOLON)
			{
				// Ensure that string is null terminated
				dataSizeText[dataPtr] = 0x00;
				
				dataLength 	= atoi(dataSizeText);
				
				if(riotServerMode == TRUE)
				{
					dataLength --;
					riotDataHandle = FALSE;
				}
				
				dataMode 	= TRUE;
				dataPtr 	= 0x00;
			}
			else if(uartData == COMMA)
			{
				dataPtr = 0x00;
			}	
			else
			{
				dataSizeText[dataPtr] = uartData;
				dataPtr ++;
			}
		}
		else
		{
			if(riotServerMode == TRUE)
			{
				if(riotDataHandle == TRUE)
				{
					if(dataPtr <= dataLength)
					{
						dataBuffer[dataPtr] = uartData;
						dataPtr ++;	
						
						if(dataLength == dataPtr)
						{
							dataAvailable = TRUE;
							handleComplete = TRUE;
							dataBuffer[dataPtr] = 0x00;
						}
					}
					else
					{
						dataAvailable = TRUE;
						handleComplete = TRUE;
						riotDataHandle = FALSE;
					}				
				}

				riotDataHandle = TRUE;
			}
			else
			{
				if(dataPtr <= dataLength)
				{
					dataBuffer[dataPtr] = uartData;
					dataPtr ++;	
					
					if(dataLength == dataPtr)
					{
						dataAvailable = TRUE;
						handleComplete = TRUE;
						dataBuffer[dataPtr] = 0x00;
					}
				}
				else
				{
					dataAvailable = TRUE;
					handleComplete = TRUE;
				}
			}
		}
	}
	// ******************************************************************************	
	


	
	// ******************************************************************************
	// Function		: commandHandle
	// Description	: Determines how to handle the command received
	// Returns		: None
	// ******************************************************************************
	void commandHandle(void)
	{
		switch(responseNumber)
		{
			case OK:
				commandInProgress 	= FALSE;
				commandSuccess		= TRUE;
				break;

			case SEND_OK:
				commandInProgress 	= FALSE;
				commandSuccess		= TRUE;
				break;

			case SEND_FAIL:
				commandInProgress 	= FALSE;
				commandSuccess		= FALSE;
				break;
				
			case SEND_ERROR:
				commandInProgress 	= FALSE;
				commandSuccess		= FALSE;
				break;	
							
			case ERROR:
				commandInProgress 	= FALSE;
				commandSuccess		= FALSE;
				break;	

			case FAIL:
				commandInProgress 	= FALSE;
				commandSuccess		= FALSE;
				break;
			/*	
			case BUSY_P:
				commandInProgress 	= FALSE;
				commandSuccess		= FALSE;
				break;
										
			case BUSY_S:
				commandInProgress 	= FALSE;
				commandSuccess		= FALSE;
				break;			
			*/

			case WIFICONNECT:
				wifiStatus = TRUE;
				break;

			case WIFICLOSED:
				wifiStatus = FALSE;
				break;
		}
	}
	// ******************************************************************************	
	


	
	// ******************************************************************************
	// Function		: tcpHandle
	// Description	: Handles connections TCP
	// Returns		: None
	// ******************************************************************************
	void tcpHandle(void)
	{
		switch(responseNumber)
		{
			case CONNECT:
				connectionStatus = TRUE;
				break;
				
			case CLOSED:
				connectionStatus = FALSE;
				break;
		}
	}
	// ******************************************************************************	
		
	
	
	// ******************************************************************************
	// Function		: timeoutChecker
	// Description	: Checks for new data, sees if it is timeout call and responds
	// Returns		: None
	// ******************************************************************************
	void timeoutChecker(void)
	{
		char respond[] = "A";
		
		if(riotServerMode == TRUE)
		{
			if(dataAvailable == TRUE)
			{
				if(dataBuffer[0] == '?')
				{
					riotDataType = RIOTTIMEOUT;
					internet_sendData(respond, 1);
				}
			}		
		}
		

	}
	// ******************************************************************************
	
		
		
	// ******************************************************************************
	// Function		: cleanup
	// Description	: Cleans up the whole system
	// Returns		: None
	// ******************************************************************************
	void cleanup(void)
	{
		unsigned char cnt;
		
		validateMode	= TRUE;
		commandMode 	= FALSE;
		dataMode 		= FALSE;
		uartData 		= FALSE;
		uartNewData 	= FALSE;
		handleComplete 	= FALSE;
		sendDataNow		= FALSE;
		ipdMode			= FALSE;
		strPointer		= 0;

		
		for(cnt = 0; cnt < NO_INSTRUCTIONS; cnt ++)
		{
			instructionStates[cnt] = TRUE;
		}		
	}
	// ******************************************************************************




	// ******************************************************************************
	// Function		: uartGet
	// Description	: This is executed when a new byte is received over UART
	// Returns		: None
	// ******************************************************************************
	void uartGet(unsigned char byteToGet)
	{
		uartData = byteToGet;
		uartNewData = TRUE;
	}
	// ******************************************************************************
	
		
	

// **********************************************************************************
